
// C++ class to manage a mail message

#ifndef _mailmsg_h 
#define _mailmsg_h

extern "C" {
#include <sys/types.h>
#include <stdio.h>
}

#include "my_regex.h"
#include "fifo.tmpl"
#include "iobottle.h"
#include "mime.h"

#define CACHED_HEADER	"CACHED HEADER"

/* Notes on the thread structure:

	All messages have next and previous pointers that point to the
	indexed messages before and after them.  They also have thread
	next and previous pointers that point to the logical messages
	before and after them, following threads.  Messages within a
	thread have "threaded" and "thread_next" set.  Messages heading
	a thread have "thread_next" set, but "threaded" is not set.
*/
#define NO_THREADS	0
#define HIDE_THREADS	1
#define SHOW_THREADS	2

/* Forward declaration mailbox data */
class mailmsg;
struct mbox_data {
	mbox_data(IObottle *mailfile);

	IObottle *MailFile;
	mailmsg *head;
	mailmsg *tail;
	int nmessages;
	int showthreads;
	bool aborted;
	bool saveall;
};

class buflen;
class mailmsg {

public:
	mailmsg(mbox_data *MBox);
	mailmsg(mbox_data *MBox, FILE *cache);
	void Add(mailmsg *msg);
	void Abort() { mbox->aborted = true; }
	bool Aborted() { return(mbox->aborted); }
	void SetSave(bool doit) { mbox->saveall = doit; }
	bool Saving(void) { return(mbox->saveall); }
	mbox_data *MBox() { return mbox; }
	~mailmsg();

	/* Mailbox stuff */
	void ThreadPass1();
	void ThreadPass2();
	int ThreadedOn(mailmsg *msg);
	void ShowThreads(int showthem) {
		mbox->showthreads = showthem;
	}
	int ShowThreads(void) {
		return(mbox->showthreads);
	}
	int NumMessages(void) {
		return(mbox->nmessages);
	}
	/* Thread navigation stuff */
	mailmsg *ByIndex(int the_index);
	mailmsg *ByStatus(const char *the_status);
	mailmsg *ByMessageID(const char *messageID);
	int ThreadSize(void);
	int ThreadIndex(void);
	mailmsg *ThreadNext(void) {
		return(thread_next);
	}
	mailmsg *RawNext(void) { return privnext; }
	mailmsg *Next(void);
	mailmsg *Next(int howmany);
	mailmsg *RawPrev(void) { return privprev; }
	mailmsg *Prev(void);
	mailmsg *Prev(int howmany);
	mailmsg *First(void) {
		return mbox->head;
	}
	mailmsg *Last(void) {
		return mbox->tail;
	}
	void SetNext(mailmsg *newnext) {
		next = newnext;
	}
	void SetPrev(mailmsg *newprev) {
		prev = newprev;
	}
	void ReOrder(void) {
		next = privnext;
		prev = privprev;
	}

	/* Header fields */
	char *GetHeader(char **keyholder);
	const char *GetField(const char *str);
	const char *MessageID(void) {
		return messageID;
	}
	char Importance(void) {
		return importance;
	}
	const char *From(void) {
		const char *from;

		from = GetField("From");
		if ( from == NULL ) {
			from = GetField("Reply-To");
		}
		return(from ? from : "Unknown");
	}
	const char *To(void) {
		const char *to;

		to = GetField("To");
		if ( to == NULL ) {
			to = GetField("Apparently-To");
		}
		if ( to == NULL ) {
			to = GetField("X-Apparently-To");
		}
		if ( to == NULL ) {
			to = GetField("Cc");
		}
		return(to ? to : "Unknown");
	}
	/* Extract a human name from an email address, if possible */
	const char *Author();
	const char *Recipient();
	/* Return the message's subject */
	const char *Subject(void) {
		const char *subject = GetField("Subject");
		return(subject ? subject : "<no subject>");
	}
	const char *Subject_Key(void) {
		return(subject_key ? subject_key : "");
	}
	/* Return the message's date in "Month Day" format */
	char *MonthDay(void);

	void SetIndex(int newindex) {
		index = newindex;
	}
	int Index(void) {
		return(index);
	}

	/* See/Set the status of the message */
	const char *Status(bool in_listing = false);
	const char *DiskStatus(void);
	void Status(const char *newstatus);
	bool Changed(void) {
		return(changed);
	}
	void NoChange(void) {
		changed = false;
	}
	mailmsg *SkipDeleted(int *nrows);
	
	/* Get the body of the message */
	MIME_body *Body(void);
	/* Make sure the body hasn't changed */
	bool Verify(void);
	/* Save the message to a file */
	int Save(char *filename);
	/* Save the status to the mailfile */
	int SaveStatus();
	/* Save the message to the mailfile */
	int SaveToDisk(Fifo<buflen> *ringbuf, int &gap);
	/* Save the header to the cache */
	int SaveToCache(FILE *cache);

	/* Get the message offset in the mailfile */
	void SetOffset(long offset) { mstart = offset; }
	void SetLength(long length) { mlength = length; }
	long Offset(void) { return mstart; }
	long Length(void) { return mlength; }

private:
	static char *boundary[];	/* The message separator */

	/* Real variables. :-) */
	int          index;
	IObottle    *MailFile;
	mbox_data   *mbox;
	const char  *messageID;
	char        *status;
	long         mstart;		/* Offset of mail message */
	long         mlength;		/* Length of message */
	char         checksum[33];	/* header checksum (text) */
	Hash<char *> fieldtab;		/* Header fields */
	MIME_body   *body;
	bool         changed;

	/* String of messages stuff */
	mailmsg *privnext, *privprev;
	mailmsg *next, *prev;
	mailmsg *thread_next, *thread_prev;
	const char *thread_key;
	char  importance;
	char *author;
	char *recipient;
	char *subject_key;
	int   subject_hash;
	int   threaded;
	int   thread_size;
	int   threaded_to;
	char *NormalizeSubject(const char *subject);
	int   HashString(char *str);

	/* Return a translation of the given status string */
	const char *TranslateStatus(const char *str) {
		if ( ! str )
			str = "N";
		else if ( *str == 'R' )
			str = " ";
		return(str);
	}
	void FinishInit();
	void ThreadMe(mailmsg *msg);
	void NewField(const char *name, const char *value);
};

#endif /* _mailmsg_h */
