
/* This is the main body of my C++ mailer */

#include <time.h>

/* Include all of the standard headers that we need */
#include "headers.h"
#include "my_regex.h"
#include "iobottle.h"
#include "terminal.h"
#include "mailmsg.h"
#include "listing.h"
#include "outgoing.h"
#include "mime.h"
#include "mimetype.h"
#include "rcfile.h"
#include "main.h"

static char *Version = 
	"My Mailer v1.6.0 by Sam Lantinga <slouken@devolution.com>";

/* The location of the user mailboxes */
#ifndef MAILPATH
//#define MAILPATH	"/var/mail"			/* A common path */
#define MAILPATH	"/usr/spool/mail"
#endif

/* A buffer structure used in a buffer ring */
class buflen {
public:
	buflen(size_t size) {
		data = new char[size];
		len = 0;
		maxlen = size;
	}
	~buflen() {
		delete[] data;
	}

public:
	char *data;
	size_t len;
	size_t maxlen;
};

/* Global variables */
IObottle *mailfile = NULL;
mailmsg  *messages = NULL;
Msg_Listing *listing = NULL;

/* Startup variables */
extern MIME_body *Startup(char *RCpath);
extern IObottle  *rcfile;
int was_compressed = 0;
int mail_locked = 0;

static bool IsSameFile(const char *a, const char *b)
{
	struct stat statA, statB;

	if ( stat(a, &statA) != 0 || stat(b, &statB) != 0 ) {
		return false;
	}
	return (statA.st_dev == statB.st_dev && statA.st_ino == statB.st_ino);
}

static void LoadStatus(mailmsg *message)
{
	static int step = 64;
	if ( step < 64 ) {
		++step;
		return;
	}
	step = 0;
	static char twiddle[] = "-+|+";
	std::cerr << "\b\b" << twiddle[message->Index()%4] << ')';
}
static void SaveStatus(mailmsg *message)
{
	static char twiddle[] = "-+|+";
	std::cerr << "\b\b" << twiddle[message->Index()%4] << ')';
}

void LoadCache(mbox_data *mbox)
{
	char cachefile[PATH_MAX];
	FILE *cache;
	mailmsg *lastmesg = NULL;

	sprintf(cachefile, "%s.cache", mailfile->Path());
	cache = fopen(cachefile, "r");
	if ( cache == NULL ) {
		return;
	}
	while ( !feof(cache) ) {
		char line[BUFSIZ];
		if ( !fgets(line, BUFSIZ, cache) ) {
			break;
		}
		if ( strcmp(line, CACHED_HEADER"\n") != 0 ) {
			break;
		}
		lastmesg = new mailmsg(mbox, cache);
		LoadStatus(lastmesg);
	}
	if ( lastmesg && lastmesg->Verify() ) {
		mailfile->seekg(lastmesg->Offset()+lastmesg->Length());

		/* Do a short read to trigger EOF */
		long here;
		char line[BUFSIZ];
		here = mailfile->tellg();
		mailfile->readline(line, BUFSIZ);
		if ( ! mailfile->eof() ) {
			mailfile->seekg(here);
		}
	} else {
		/* Cache failed, oh well, load normally */
		mailmsg *current, *next;
		current = next = mbox->head;
		while ( next ) {
			next = current->RawNext();
			delete current;
			current = next;
		}
		mbox->head = mbox->tail = NULL;
	}
	fclose(cache);
}

void SaveCache()
{
	char cachefile[PATH_MAX];
	FILE *cache;
	mailmsg *lastmesg;

	sprintf(cachefile, "%s.cache", mailfile->Path());
	cache = fopen(cachefile, "w");
	if ( cache == NULL ) {
		return;
	}
	/* Need to fix up the indices so cached thread indexes are correct */
	int index = 1;
	for ( lastmesg = messages; lastmesg; lastmesg = lastmesg->RawNext() ) {
		if ( *lastmesg->Status() != 'D' ) {
			lastmesg->SetIndex(index++);
		}
	}
	for ( lastmesg = messages; lastmesg; lastmesg = lastmesg->RawNext() ) {
		if ( *lastmesg->Status() != 'D' ) {
			lastmesg->SaveToCache(cache);
		}
	}
	fclose(cache);
}

void CleanupMessages(void)
{
	mailmsg *current, *next;

	/* Check to see if our status changed */
	if ( messages->Saving() ) {
		char *trashpath = ExpandPath(GetStartVar("TrashFile"));

		/* Make sure we're not actually working on the trash */
		if ( trashpath && IsSameFile(trashpath, mailfile->Path()) ) {
			delete[] trashpath;
			trashpath = NULL;
		}

		/* Skip messages that don't need saving */
		for ( current=messages;
		      current;
		      current = current->RawNext() ) {
			if ( *current->Status() == *current->DiskStatus() ) {
				continue;
			}
			if ( *current->Status() == 'D' && !current->Aborted() ) {
				/* Need to start moving messages around */
				break;
			}
			bool saved;
			MailLock();
			saved = (current->SaveStatus() == 0);
			MailUnlock();
			if ( ! saved ) {
				/* Need to start saving to disk */
				break;
			}
		}
		if ( current != NULL ) {
			std::cout << "Saving  messages: ( )" << std::flush;

			/* Lock the mailfile */
			MailLock();

			Fifo<buflen> *ringbuf = new Fifo<buflen>;
			mailfile->clrerr();
			long offset = current->Offset();
			mailfile->seekg(offset);
			mailfile->seekp(offset);

			/* Delete the messages */
			int gap = 0;
			while ( current ) {
				long mlength;
				if ( *current->Status() == 'D' && trashpath ) {
					current->Save(trashpath);
				}
				mlength = current->SaveToDisk(ringbuf, gap);
				current->SetOffset(offset);
				current->SetLength(mlength);
				offset += mlength;
				SaveStatus(current);
				current = current->RawNext();
				gap = FlushRing(ringbuf, gap);
			}

			/* New Mail! */
			int newmail = 0;
			int newlen;
			char newdata[BUFSIZ];
			while ( (newlen=mailfile->read(newdata, BUFSIZ)) > 0 ) {
				QueueRing(ringbuf, newdata, newlen);
				gap -= newlen;
				gap = FlushRing(ringbuf, gap);
				++newmail;
			}
			/* Flush the ring -- write past EOF here */
			buflen *bptr;
			while ( (bptr=ringbuf->Pull()) ) {
				if ( mailfile->write(bptr->data, bptr->len) !=
								bptr->len ) {
				/* FIXME!  Abort -- disk full? */
				}
				delete bptr;
			}
			mailfile->flush();
			mailfile->truncate(mailfile->tellp());

			time_t now = time(NULL);
			if ( newmail )
				mailfile->set_time(now-1, now);
			else
				mailfile->set_time(now, now-1);

			/* Unlock the mailfile */
			MailUnlock();

			std::cout << " Done." <<  std::endl;
		}
		SaveCache();

		if ( trashpath != NULL ) {
			delete[] trashpath;
		}
	}

	/* Delete the messages */
	delete messages->MBox();
	current = next = messages;
	while ( next ) {
		next = current->RawNext();
		delete current;
		current = next;
	}
}

/* Note that it's possible for us to be called multiple times:
	e.g. if there's a segmentation violation from within this function..
*/
void quit(int exitcode)
{
	/* Ignore signals while we clean up */
	signal(SIGHUP, SIG_IGN);
	signal(SIGINT, SIG_IGN);
	signal(SIGQUIT, SIG_IGN);
	signal(SIGSEGV, SIG_DFL);
	signal(SIGBUS, SIG_DFL);
	signal(SIGTERM, SIG_IGN);
#ifdef SIGWINCH
	signal(SIGWINCH, SIG_DFL);
#endif

	/* Clean up everything */
	if ( listing ) {
		delete listing;
		listing = NULL;
	}

	/* Why are we exiting? */
	switch(exitcode) {
		case SIGHUP:
			if ( messages )
				messages->Abort();
			std::cout << "\rExiting because of hang-up signal!" << std::endl;
			break;
		case SIGINT:
		case SIGQUIT:
			if ( messages )
				messages->Abort();
			std::cout << "\rExiting because of keyboard interrupt!" <<
									std::endl;
			break;
		case SIGSEGV:
		case SIGBUS:
			if ( messages )
				messages->Abort();
			std::cout << "\rExiting because of program bug!" << std::endl;
			break;
		case SIGTERM:
			if ( messages )
				messages->Abort();
			std::cout << "\rExiting because of termination signal!" <<
									std::endl;
			break;
		default:
			break;
	}
	/* Make sure we have a mailfile and messages */
	if ( mailfile ) {
		char cmdbuf[BUFSIZ];

		MailUnlock();
		if ( messages ) {
			CleanupMessages();
			messages = NULL;

			///* Check for new mail */
			//time_t atime, mtime;
			//mailfile->get_time(&atime, &mtime);
			//if ( mtime > atime )
			//	std::cout << "You have NEW mail!" << std::endl;
		}
		if ( was_compressed ) {
			/* For some reason, this crashes on some mailboxes
				cmdbuf = new char[strlen("gzip "),
						strlen(mailfile->Path())+1];
			*/
			sprintf(cmdbuf, "gzip %s", mailfile->Path());
		}
		delete mailfile;
		mailfile = NULL;

		/* Any final mailfile cleanup */
		if ( was_compressed ) {
			std::cout << "Compressing mailfile..." << std::flush;
			(void) system(cmdbuf);
			std::cout << " Done." << std::endl;
			/* For some reason, this crashes on some mailboxes
				delete[] cmdbuf;
			*/
		}
	}
	Quit_MIMEtypes();
	if ( startup ) {
		delete startup;
		delete rcfile;
		startup = NULL;
	}
	exit(exitcode);
}

/* Notify the listing of a window size change */
void resize(int sig)
{
	signal(sig, resize);
	if ( listing ) {
		unsetenv("LINES");
		unsetenv("COLUMNS");
		listing->ResizePending();
	}
}

void QueueRing(Fifo<buflen> *ringbuf, char *data, int len)
{
	buflen *bptr;

	if ( len > 0 ) {
		bptr = new buflen(len);
		bptr->len = len;
		memcpy(bptr->data, data, len);
		/* Put back the newline */
		if ( data[len-1] == '\0' )
			bptr->data[len-1] = '\n';
		ringbuf->Push(bptr);
	}
}

/* Note that the gap ranges from 0 to negative infinity */
int FlushRing(Fifo<buflen> *ringbuf, int gap)
{
	buflen *bptr;

	/* See if we can write out a bunch of data */
	while ( (bptr=ringbuf->Peek()) && ((bptr->len+gap) <= 0) ) {
		if ( mailfile->write(bptr->data, bptr->len) != bptr->len ) {
			/* FIXME!  Abort -- disk full? */
		}
		gap += bptr->len;
		ringbuf->Pull();
		delete bptr;
	}
	return(gap);
}

void MailLock(void)
{
	char *lockfile;
	int   lockfd;

	/* Figure out the lock file */
	lockfile = new char[strlen(mailfile->Path())+5+1];
	sprintf(lockfile, "%s.lock", mailfile->Path());

	/* Assume a blocking Lock() call */
	while ( ((lockfd=open(lockfile, (O_WRONLY|O_CREAT|O_EXCL), 0)) < 0) &&
							(errno == EEXIST) ) {
		sleep(1);
	}
	if ( lockfd >= 0 )
		close(lockfd);
	delete[] lockfile;

	/* The lock failing could be a problem.. */
	(void) mailfile->lock();		/* Blocking call */

	mail_locked = 1;
}

void MailUnlock(void)
{
	/* Figure out the lock file */
	if ( mail_locked ) {
		char *lockfile;

		lockfile = new char[strlen(mailfile->Path())+5+1];
		sprintf(lockfile, "%s.lock", mailfile->Path());
		(void) unlink(lockfile);
		delete[] lockfile;

		(void) mailfile->unlock();
	}
}

int CheckLock(const char *path)
{
	struct stat sb;
	char *lockfile;
	int   retval;

	/* Just try to stat() the lockfile */
	lockfile = new char[strlen(path)+5+1];
	sprintf(lockfile, "%s.lock", path);
	retval = stat(lockfile, &sb);
	delete[] lockfile;
	return(retval == 0 ? 1 : 0);
}

static time_t last_mtime = 0;
static off_t  last_size  = 0;
mailmsg *CheckMail(mailmsg *current, Terminal *tty, int *changed)
{
	time_t   mtime;
	off_t    size;
	mailmsg *newmsgs, *lastmesg;
	mailmsg *origptr, *otherptr;
	int      oldwait, answer;
	char     buffer[BUFSIZ];

	/* Check for modified mailbox */
	mailfile->get_time(NULL, &mtime);
	mailfile->Size(&size);
	if ( (mtime == last_mtime) && (size == last_size) ) {
		return(current);
	}
	last_mtime = mtime;
	last_size = size;

	/* If we are aborted here, we don't save the mailfile */
	messages->SetSave(false);

	/* If last message is fine, assume newly arrived mail */
	MailLock();
	lastmesg = messages->Last();
	if ( lastmesg->Verify() ) {
		tty->status(0, "You have NEW mail!");
		sleep(1);

		tty->status(0, "Loading messages: ( )");
		mailfile->seekg(lastmesg->Offset()+lastmesg->Length());
		while ( ! mailfile->eof() ) {
			lastmesg = new mailmsg(messages->MBox());
			lastmesg->ThreadPass1();
			lastmesg->ThreadPass2();
			LoadStatus(lastmesg);
		}
		MailUnlock();

		messages->SetSave(true);
		if ( changed )
			++(*changed);
		return(current);
	} else {
		MailUnlock();
	}

	oldwait = tty->setwait(5*2);	/* Wait for 5 seconds */
	answer = tty->promptchar(0,
		"Mailfile has changed.  [R]eload messages or [q]uit? ");
	tty->setwait(oldwait);
	if ( tolower(answer) == 'q' )
		return(NULL);
		
	/* Make sure there is mail there */
	mailfile->seekg(0);
	if ( mailfile->readline(buffer, BUFSIZ) <= 1 ) {
		tty->status(0, "No mail left in %s", mailfile->Path());
		sleep(3);
		return(NULL);
	}
	mailfile->seekg(0);

	/* Create a new set of messages */
	mbox_data *mbox = new mbox_data(mailfile);
	MailLock();
	tty->status(0, "Reloading messages: ( )");
	while ( ! mailfile->eof() ) {
		lastmesg = new mailmsg(mbox);
		LoadStatus(lastmesg);
	}
	MailUnlock();
	tty->status(0, "Threading messages...");
	for ( lastmesg = mbox->head; lastmesg; lastmesg = lastmesg->RawNext() ) {
		lastmesg->ThreadPass1();
	}
	for ( lastmesg = mbox->head; lastmesg; lastmesg = lastmesg->RawNext() ) {
		lastmesg->ThreadPass2();
	}
	newmsgs = mbox->head;
	newmsgs->ShowThreads(messages->ShowThreads());

	/* Mark deleted messages as deleted */
	for ( origptr = messages; origptr; origptr = origptr->RawNext() ) {
		if ( *origptr->Status() == *origptr->DiskStatus() )
			continue;

		otherptr = newmsgs->ByMessageID(origptr->MessageID());
		if ( otherptr != NULL )
			otherptr->Status(origptr->Status());
		LoadStatus(origptr);
	}

	/* Try to keep the same current message */
	current = newmsgs->ByMessageID(current->MessageID());
	if ( ! current ) {
		if ( ((current = newmsgs->ByStatus("N")) == NULL) &&
		     ((current = newmsgs->ByStatus("O")) == NULL) ) {
			current = mbox->tail;
		}
	}

	/* Now delete the old messages */
	CleanupMessages();
	messages = newmsgs;

	/* Return. :) */
	if ( changed )
		++(*changed);
	return(current);
}

/* cuserid() has been obsoleted */
static char *mycuserid(void)
{
	struct passwd *pwd;

	pwd = getpwuid(geteuid());
	if ( pwd == NULL )
		return(NULL);
	return(pwd->pw_name);
}

void Usage(char *prog)
{
	std::cout << 
Version << std::endl << std::endl <<
"Usage: " << prog << " [-s subject] recipient1 recipient2 ... " << std::endl <<
"or...  " << prog << " -c" << std::endl <<
"or...  " << prog << " [-r initfile] [-f mailbox]" << std::endl <<
	std::endl;
}

int main(int argc, char *argv[])
{
	/* Flags */
	int   check_mail = 0;
	char *rcpath = NULL;
	char *subject = NULL;

	/* Normal variables */
	char *mailpath = NULL;
	char *whoami;
	char *variable;
	mailmsg *current;

	/* Parse the command line arguments */
	int arg;
	while ( (arg=getopt(argc, argv, "cf:hr:s:v")) != EOF ) {
		switch (arg) {
			case 'c':
				/* Check for new mail */
				check_mail = 1;
				break;
			case 'h':
				Usage(argv[0]);
				quit(0);
			case 'f':
				mailpath = optarg;
				break;
			case 'r':
				rcpath = optarg;
				break;
			case 's':
				subject = optarg;
				break;
			case 'v':
				/* Verbose flag -- 
				   It's hard to show sendmail output in curses.
				 */
				break;
			default:
				Usage(argv[0]);
				quit(0);
		}
	}
	argc -= optind;
	argv += optind;
	umask(066);			/* Protect our tempfiles */

	/* Open the startup file */
	startup = Startup(rcpath);

	/* Load MIME type database */
	Init_MIMEtypes(NULL);

	/* Command line mailing? */
	if ( *argv ) {
		if ( NewMail(argv, subject) < 0 ) {
			std::cerr << "Mail send failed!" << std::endl;
			quit(-1);
		}
		quit(0);
	}

	/* Use the default mailfile, if defined */
	if ( mailpath == NULL )
		mailpath = GetStartVar("MailFile");

	/* Find out who we are and what our mailfile is */
	if ( (mailpath == NULL) &&
			((mailpath=(char *)getenv("MAIL")) == NULL) ) {
		if ( ((whoami=getlogin()) == NULL) &&
					((whoami=mycuserid()) == NULL) ) {
			std::cerr << "Who are you?  -- Aborting!" << std::endl;
			quit(-1);
		}
		mailpath = new char[strlen(MAILPATH)+1+strlen(whoami)+1];
		sprintf(mailpath, "%s/%s", MAILPATH, whoami);
	}
	/* Perform tilde expansion on the mail file name */
	mailpath = ExpandPath(mailpath);

	/* Check for new mail and quit */
	struct stat sb;

	if ( stat(mailpath, &sb) < 0 ) {
		std::cout << mailpath << " is not readable" << std::endl;
		quit(0);
	}
	if ( sb.st_size == 0 )  {
		std::cout << "No mail." << std::endl;
		quit(0);
	}
	if ( check_mail ) {
		if ( sb.st_mtime < sb.st_atime )
			std::cout << "You have mail." << std::endl;
		else
			std::cout << "You have NEW mail." << std::endl;
		quit(0);
	}

	/* Check to see if the lock file exists */
	if ( CheckLock(mailpath) ) {
		std::cout << "Lockfile (" << mailpath << ".lock) exists!" << std::endl;
		std::cout << "Please try again later..." << std::endl;
		quit(0);
	}

	/* Check for compressed mail */
	{
		struct stat sb;
		char *ptr, *cmdbuf = new
				char[strlen("gzip -d ")+strlen(mailpath)+3+1];
		int decompress = 0;

		/* Check for compressed mailfile */
		sprintf(cmdbuf, "gzip -d %s", mailpath);
		ptr = cmdbuf + strlen("gzip -d ")+strlen(mailpath)-3;
		if ( strcmp(ptr, ".gz") == 0 ) {
			/* Strip .gz ending on mailpath */
			ptr = mailpath+strlen(mailpath)-3;
			*ptr = '\0';
			decompress = 1;
		} else if ( stat(ptr, &sb) < 0 ) {
			strcat(cmdbuf, ".gz");
			ptr = cmdbuf + strlen("gzip -d ");
			if ( stat(ptr, &sb) == 0 )
				decompress = 1;
		}

		/* Perform decompression */
		if ( decompress ) {
			std::cout << "Uncompressing mailfile..." << std::flush;
			if ( system(cmdbuf) == 0 )
				was_compressed = 1;
			std::cout << " Done." << std::endl;
		}
		delete[] cmdbuf;
	}
		
	/* Set up the signal handlers */
	signal(SIGHUP, quit);
	signal(SIGINT, quit);
	signal(SIGQUIT, quit);
	signal(SIGSEGV, quit);
	signal(SIGBUS, quit);
	signal(SIGTERM, quit);
#ifdef SIGWINCH
	signal(SIGWINCH, resize);
#endif

	/* If we get here, we are reading mail */
	mailfile = new IObottle(mailpath);
	char             buffer[BUFSIZ];

	if ( mailfile->fail() || mailfile->eof() ) {
		std::cout << "No mail in " << mailpath << std::endl;
		quit(0);
	}
	if ( mailfile->readline(buffer, BUFSIZ) == 1 ) {
		std::cout << "Newline at start of mailfile -- aborting!" << std::endl;
		quit(-1);
	}
	mailfile->seekg(0);

	/* Set the MIME native character set */
	MIME_body::MIME_Charset(GetStartVar("MIME_charset"));

	/* Set the MIME content types to ignore */
	MIME_body::MIME_Ignore(GetStartVar("MIME_ignore"));

	/* If we have succeeded in opening the mailfile, load mail */
	mailmsg *lastmesg;
	mbox_data *mbox = new mbox_data(mailfile);
	MailLock();
	std::cout << "Loading messages: ( )" << std::flush;
	LoadCache(mbox);
	while ( ! mailfile->eof() ) {
		lastmesg = new mailmsg(mbox);
		LoadStatus(lastmesg);
	}
	MailUnlock();
	std::cout << "\rThreading messages..." << std::flush;
	for ( lastmesg = mbox->head; lastmesg; lastmesg = lastmesg->RawNext() ) {
		lastmesg->ThreadPass1();
	}
	for ( lastmesg = mbox->head; lastmesg; lastmesg = lastmesg->RawNext() ) {
		lastmesg->ThreadPass2();
	}
	messages = mbox->head;
	if ( ((current = messages->ByStatus("N")) == NULL) &&
	     ((current = messages->ByStatus("O")) == NULL) ) {
		current = mbox->tail;
	}
	std::cout << " Done." <<  std::endl;
	sleep(1);
	mailfile->get_time(NULL, &last_mtime);
	mailfile->Size(&last_size);

	/* Set the threading, if desired */
	if ( (variable=GetStartVar("Threading")) != NULL ) {
		if ( strcmp(variable, "NO_THREADS") == 0 )
			messages->ShowThreads(NO_THREADS);
		else
		if ( strcmp(variable, "HIDE_THREADS") == 0 )
			messages->ShowThreads(HIDE_THREADS);
		else
		if ( strcmp(variable, "SHOW_THREADS") == 0 )
			messages->ShowThreads(SHOW_THREADS);
	}

	/* Note that the listing will control the new "current" message */
	listing = new Msg_Listing(mailfile, current);

	/* Set whether we're looking at sent mail */
	char *sentfile = ExpandPath(GetStartVar("SentFile"));
	if ( sentfile ) {
		listing->SetSentMode(IsSameFile(sentfile, mailfile->Path()));
		delete[] sentfile;
	}

	/* Set the listing format */
	if ( (variable=GetStartVar("IndentThread")) != NULL ) {
		if ( tolower(*variable) == 'n' )
			listing->IndentThread(0);
		else
			listing->IndentThread(1);
	}

	/* Main command loop */
	listing->RunMain(CheckMail);
	quit(0);
}
